/* -------------------------------------------------------------------------
   Copyright (c) 1998 - 2001 enGenic.  All rights reserved.
   ------------------------------------------------------------------------- */
#include "..\escutil\iv_str.h"
#include "..\escutil\varcontroller.h"
#include "odbc_db.h"

enum DSN_STRING { DSN_NONE, DSN_DSN, DSN_SRVR, DSN_UID, DSN_PWD };

void DisplayError (
   Log        *ivLog,
   SQLRETURN   nResult,
   SWORD       fHandleType,
   SQLHANDLE   handle);

#define EATERRORS DisplayError

void EatErrors (
   Log        *ivLog,
   SQLRETURN   nResult,
   SWORD       fHandleType,
   SQLHANDLE   handle);

// *************************************************************************
// ********************** ODBC Manager Class *******************************
// *************************************************************************

ODBCManager::ODBCManager(Log* outLog)
{
   ivLog = outLog;
}

ODBCManager::~ODBCManager(void)
{
   // Close all connections
   Reset();
}

SQLList* ODBCManager::FindConnection(const char* connectionHandle)
{
   string sRequestedHandle;
   char* value;

   if(connectionHandle)
   {
      // Trim, copy to a string, and make all uppercase characters
      value = TrimStr(connectionHandle);
      sRequestedHandle = value;
      delete[] value;
      transform(sRequestedHandle.begin(), sRequestedHandle.end(), sRequestedHandle.begin(), toupper);

      if(ConnectionsList.count(sRequestedHandle) > 0)
      {
         return(ConnectionsList[sRequestedHandle]);
      }
   }
   // else
   return(NULL);
}

IV_STATUS ODBCManager::Connect(const char* connectHandle, const char* connectParms)
{
   string sNewHandle;
   IV_STATUS status;
   SQLList* pNewSQLList;
   char* value;

   if(connectHandle == NULL)
   {
      ivLog->Print("      ERROR: No connection handle supplied\n");
      return(IVR_FATAL);
   }

   // Trim, Copy to a string, and make all uppercase characters
   value = TrimStr(connectHandle);
   sNewHandle = value;
   delete[] value;
   transform(sNewHandle.begin(), sNewHandle.end(), sNewHandle.begin(), toupper);

   // Make a new SQL list object
   pNewSQLList = new SQLList(ivLog);
   if(pNewSQLList)
   {
      // Use it to connect with
      status = pNewSQLList->Connect(connectParms);

      if(status == IVR_SUCCESS)
      {
         // Store it in the ConnectionList map
         ConnectionsList[sNewHandle] = pNewSQLList;
      }
      else
      {
         delete pNewSQLList;
      }

      return(status);
   }
   else
   {
      ivLog->Print("      ERROR: Unable to allocate a new SQL connection\n");
      return(IVR_FATAL);
   }
}

IV_STATUS ODBCManager::Disconnect(const char* connectHandle)
{
   SQLList *connection;
   string sHandle;
   char* value;

   if(connectHandle == NULL)
   {
      // shouldn't happen, but to be safe, if no handle supplied, just exit
      return(IVR_SUCCESS);
   }

   // Find ODBC connection by its handle
   if((connection = FindConnection(connectHandle)) == NULL)
   {
      // Not found means its as closed as it needs to be!
      return(IVR_SUCCESS);
   }

   // Delete SQLList object (destructor performs disconnect from data source)
   delete connection;

   // Erase element from the 'ConnectionsList' map
   // requires trimmed, uppercase string as element key
   value = TrimStr(connectHandle);
   sHandle = value;
   delete[] value;
   value = NULL;
   transform(sHandle.begin(), sHandle.end(), sHandle.begin(), toupper);

   ConnectionsList.erase(sHandle);

   // Note: ConnectionList contains a pointer to SQLList object only,
   //       so this code will not cause a double-delete

   return(IVR_SUCCESS);
}

void ODBCManager::Reset(void)
{
   map<string,SQLList*>::iterator i;

   for(i = ConnectionsList.begin(); i != ConnectionsList.end(); i++)
   {
      // Delete SQLList object (destructor performs disconnect from data source)
      delete (i->second);
   }
   // Remove all elements in 'ConnectionsList' map
   ConnectionsList.clear();
}

/* *************************************************************************
   ********************** SQLList Class ************************************
   ************************************************************************* */

/* -------------------------------------------------------------------------
   Name:        SQLList::SQLList                               (Constructor)
   ------------------------------------------------------------------------- */
SQLList::SQLList(Log* outLog) : Head()
{
   ivLog = outLog;
   henv = SQL_NULL_HANDLE;
   hdbc = SQL_NULL_HANDLE;

   Reset();
}
/*SQLList::SQLList*/

/* -------------------------------------------------------------------------
   Name:        SQLList::~SQLList                               (Destructor)
   ------------------------------------------------------------------------- */
SQLList::~SQLList(void)
{
   Clear();
   Disconnect();

   if(henv != SQL_NULL_HANDLE)
   {
      SQLFreeHandle(SQL_HANDLE_ENV, henv);
   }
}
/*SQLList::~SQLList*/

/* -------------------------------------------------------------------------
   Name:        SQLList::Connect

   Purpose:     Connect to a standard database (non-SQL).

   Notes:       - Use SQLGetInfo with SQL_TXN_CAPABLE to determine if driver
                  is capable of performing transactions.

   Notes:       - Look up SQLDriverConnect.  This may be what is needed to
                  connect to Oracle.
   ------------------------------------------------------------------------- */
IV_STATUS SQLList::Connect(const char* connectString)
{
   SQLRETURN    ret;
   SQLCHAR      pszOutConnectStr[1024];
   SQLSMALLINT  iOutStrLen;
   SQLINTEGER   oldConnTimeout;
   SQLINTEGER   newConnTimeout;
   const char*  pszAfterTimeout;
   bool         bSetTimeout = false;

	// Sean change to protect against exception errors when connecting with an SQL
   // handle that is already in use
   Disconnect();

   // Store original connection string for later comparisons
   SaveConnectionString(connectString);

   pszAfterTimeout = connectString;
   if(strncmp(pszAfterTimeout, "TIMEOUT=", 8) == 0)
   {
      bSetTimeout = true;
      pszAfterTimeout += 8;

      newConnTimeout = atoi (pszAfterTimeout);

      pszAfterTimeout = strchr (connectString, ';');
      if(pszAfterTimeout == NULL)
      {
         pszAfterTimeout = strchr(connectString, '\0');
      }
      else
      {
         pszAfterTimeout++;
      }
   }

   // Allocate a new SQLHDBC. If successful then call SQLConnect using the
   // allocated hdbc and supplied connection information.
   ret = SQLAllocHandle (
            SQL_HANDLE_DBC,             // SQLSMALLINT  HandleType
            henv,                       // SQLHANDLE    InputHandle
            &hdbc);                     // SQLHANDLE *  OutputHandle
   if(ret != SQL_SUCCESS)
   {
      ivLog->Print ("ERROR: Connect: Attempt to allocate handle failed.\n");
      DisplayError (ivLog, ret, SQL_HANDLE_DBC, henv);
      return IVR_FAIL;
   }

   if(bSetTimeout)
   {
      ret = SQLGetConnectAttr (
               hdbc,                       // SQLHDBC        ConnectionHandle
               SQL_ATTR_LOGIN_TIMEOUT,     // SQLINTEGER     Attribute
               &oldConnTimeout,            // SQLPOINTER     ValuePtr
               0,                          // SQLINTEGER     BufferLength
               0);                         // SQLINTEGER    *StringLengthPtr

      if((ret != SQL_SUCCESS) && (ret != SQL_SUCCESS_WITH_INFO))
      {
         ivLog->Print ("Failed to get connection timeout.\n");
         DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
         SQLFreeHandle (SQL_HANDLE_DBC, hdbc);
      }
      else if(ret == SQL_SUCCESS_WITH_INFO)
      {
         ivLog->Print ("Default connection timeout: %d\n", oldConnTimeout);
         DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
      }
      else
      {
         ivLog->Print ("Default connection timeout: %d\n", oldConnTimeout);
      }

      ret = SQLSetConnectAttr (
               hdbc,                       // SQLHDBC        ConnectionHandle
               SQL_ATTR_LOGIN_TIMEOUT,     // SQLINTEGER     Attribute
               (SQLPOINTER)newConnTimeout, // SQLPOINTER     ValuePtr
               0);                         // SQLINTEGER     StringLength

      if((ret != SQL_SUCCESS) && (ret != SQL_SUCCESS_WITH_INFO))
      {
         ivLog->Print ("Failed to set connection timeout to %d seconds.\n",
                       newConnTimeout);
         DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
         SQLFreeHandle (SQL_HANDLE_DBC, hdbc);
      }
      else if(ret == SQL_SUCCESS_WITH_INFO)
      {
         ivLog->Print ("Connection timeout set to %d seconds.\n",
                       newConnTimeout);
         DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
      }
      else
      {
         ivLog->Print ("Connection timeout set to %d seconds.\n", newConnTimeout);
      }
   }

   ret = SQLDriverConnect (
            hdbc,                       // SQLHDBC        ConnectionHandle
            NULL,                       // SQLHWND        WindowHandle
            (SQLCHAR*)pszAfterTimeout,  // SQLCHAR *      InConnectionString
            (SQLSMALLINT)
            (strlen(pszAfterTimeout)+1),// SQLSMALLINT    StringLen
            (SQLCHAR*)pszOutConnectStr, // SQLCHAR *      OutConnectionString
            1024,                       // SQLSMALLINT    BufferLength
            &iOutStrLen,                // SQLSMALLINT *  StringLength2Ptr
            SQL_DRIVER_NOPROMPT);       // SQLUSMALLINT   DriverCompletion

   // if failed to connect, free the allocated hdbc before return
   if((ret != SQL_SUCCESS) && (ret != SQL_SUCCESS_WITH_INFO))
   {
      ivLog->Print ("ERROR: Connect: SQLConnect failed.\n");
      DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
      SQLFreeHandle (SQL_HANDLE_DBC, hdbc);
      hdbc = SQL_NULL_HANDLE;
      return IVR_FAIL;
   }
// ivLog->Print ("Connect string: '%s'\n", pszOutConnectStr);

   // Display information if driver returns SQL_SUCCESS_WITH_INFO.
   if(ret == SQL_SUCCESS_WITH_INFO)
   {
      EATERRORS (ivLog, ret, SQL_HANDLE_DBC, hdbc);
   }

   // -------------------------------------------------------------------------
   // These two calls will fail if not using an Intersolv driver.
   // -------------------------------------------------------------------------
   ret = SQLSetConnectAttr (hdbc, 1041, "ivee.lic", SQL_NTS);
   if((ret == SQL_SUCCESS) || (ret == SQL_SUCCESS_WITH_INFO))
   {
      if(ret == SQL_SUCCESS_WITH_INFO)
      {
         EATERRORS (ivLog, ret, SQL_HANDLE_DBC, hdbc);
      }

      ret = SQLSetConnectAttr (hdbc, 1042, "T43de34qj778dfe4332ft556", SQL_NTS);
      if(ret != SQL_SUCCESS)
      {
         DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
      }
   }
   else
   {
      EATERRORS (ivLog, ret, SQL_HANDLE_DBC, hdbc);
   }

   return(IVR_SUCCESS);
}
/*SQLList::Connect*/

SQLHDBC SQLList::GetDBHandle(void)
{
   return(hdbc);
}

bool SQLList::Okay(void)
{
   return(okay);
}

void SQLList::Reset(void)
{
   Clear();
   Disconnect();

   if(henv != SQL_NULL_HANDLE)
   {
      SQLFreeHandle(SQL_HANDLE_ENV, henv);
      henv = SQL_NULL_HANDLE;
   }

   hdbc = SQL_NULL_HANDLE;

   if(SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &henv) != SQL_SUCCESS)
   {
      ivLog->Print("ERROR: Attempt to allocate SQL handle failed.\n");
      okay = false;
      return;
   }

   if(SQLSetEnvAttr(henv, SQL_ATTR_ODBC_VERSION, (SQLPOINTER) SQL_OV_ODBC3, SQL_IS_INTEGER) != SQL_SUCCESS)
   {
      ivLog->Print ("ERROR: Failed to set SQL environment attribute.\n");
      okay = false;
      return;
   }

   okay = true;
}

/* -------------------------------------------------------------------------
   Name:        SQLList::DetermineTransCapability

   Purpose:     May need to determine the transaction capabilities of the
                database before calling transaction functions.
   ------------------------------------------------------------------------- */
IV_STATUS SQLList::DetermineTransCapability(void)
{
   SQLRETURN    ret;
   SQLUSMALLINT transSupported;
   SQLUINTEGER  transIsolation;
   SQLUINTEGER  defaultTransIsolation;

   ret = SQLGetInfo (
            hdbc,                       // SQLHDBC      ConnectionHandle
            SQL_TXN_CAPABLE,            // SQLUSMALLINT InfoType
            &transSupported,            // SQLPOINTER   InfoValuePtr
            sizeof(SQLUSMALLINT),       // SQLSMALLINT  BufferLength
            NULL);                      // SQLSMALLINT *StringLengthPtr
   if (ret == SQL_SUCCESS) {
      char  *s;

      switch (transSupported) {
         case SQL_TC_DML:
            s = "dml only; ddl statements in transaction cause an error";
            break;
         case SQL_TC_DDL_COMMIT:
            s = "ddl in transaction cause commit";
            break;
         case SQL_TC_DDL_IGNORE:
            s = "ddl statements ignored during transaction";
            break;
         case SQL_TC_ALL:
            s = "transaction can contain both dml and ddl in any order";
            break;
         case SQL_TC_NONE:
            break;
         default:
            s = "not supported";
      }
      ivLog->Print ("         Transaction support: %s\n", s);

   } else {
      ivLog->Print ("Error occured trying to determine transaction support.\n");
      DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
   }

   ret = SQLGetInfo (
            hdbc,                       // SQLHDBC      ConnectionHandle
            SQL_DEFAULT_TXN_ISOLATION,  // SQLUSMALLINT InfoType
            &defaultTransIsolation,     // SQLPOINTER   InfoValuePtr
            sizeof(SQLUINTEGER),        // SQLSMALLINT  BufferLength
            NULL);                      // SQLSMALLINT *StringLengthPtr
   if (ret == SQL_SUCCESS) {
      if (defaultTransIsolation & SQL_TXN_READ_UNCOMMITTED)
         ivLog->Print ("         Default isolation level: Read Uncommitted\n");
      if (defaultTransIsolation & SQL_TXN_READ_COMMITTED)
         ivLog->Print ("         Default isolation level: Read committed\n");
      if (defaultTransIsolation & SQL_TXN_REPEATABLE_READ)
         ivLog->Print ("         Default isolation level: Repeatable\n");
      if (defaultTransIsolation & SQL_TXN_SERIALIZABLE)
         ivLog->Print ("         Default isolation level: Serializable\n");

   } else {
      ivLog->Print ("Error occured trying to determine isolation.\n");
      DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
   }

   ret = SQLGetInfo (
            hdbc,                       // SQLHDBC      ConnectionHandle
            SQL_TXN_ISOLATION_OPTION,   // SQLUSMALLINT InfoType
            &transIsolation,            // SQLPOINTER   InfoValuePtr
            sizeof(SQLUINTEGER),        // SQLSMALLINT  BufferLength
            NULL);                      // SQLSMALLINT *StringLengthPtr
   if (ret == SQL_SUCCESS) {
      if (transIsolation & SQL_TXN_READ_UNCOMMITTED)
         ivLog->Print ("         Available isolation level: Read Uncommitted\n");
      if (transIsolation & SQL_TXN_READ_COMMITTED)
         ivLog->Print ("         Available isolation level: Read committed\n");
      if (transIsolation & SQL_TXN_REPEATABLE_READ)
         ivLog->Print ("         Available isolation level: Repeatable\n");
      if (transIsolation & SQL_TXN_SERIALIZABLE)
         ivLog->Print ("         Available isolation level: Serializable\n");

   } else {
      ivLog->Print ("Error occured trying to determine isolation option.\n");
      DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
   }

   return IVR_SUCCESS;
}
/*SQLList::DetermineTransCapability*/

/* -------------------------------------------------------------------------
   Name:        SQLList::Disconnect

   Purpose:     Disconnect from the data source.  Close any transactions
                that are open *without* committing data.  Cursors are
                automatically closed when the database is closed.
   ------------------------------------------------------------------------- */
IV_STATUS SQLList::Disconnect(void)
{
   // call SQLDisconnect and SQLFreeConnect to disconnect and free hdbc
   if (hdbc != SQL_NULL_HANDLE) {
      SQLDisconnect (hdbc);
      SQLFreeHandle (SQL_HANDLE_DBC, hdbc);
      hdbc = SQL_NULL_HANDLE;
      SetChildDBHandle();
   }

   return IVR_SUCCESS;
}
/*SQLList::Disconnect*/

/* -------------------------------------------------------------------------
   Name:        SQLList::BeginTrans

   Purpose:     Starts a transaction.  Changes will not be committed until
                'EndTrans' is called.

   Notes:       - Different databases can support different isolation
                  levels.
   ------------------------------------------------------------------------- */
IV_STATUS SQLList::BeginTrans(void)
{
   SQLRETURN  ret;

   if (transOpen)
      return IVR_SUCCESS;

   ret = SQLSetConnectAttr (
            hdbc,                             // SQLHDBC    ConnectionHandle
            SQL_ATTR_AUTOCOMMIT,              // SQLINTEGER Attribute
            (SQLPOINTER) SQL_AUTOCOMMIT_OFF,  // SQLPOINTER ValuePtr
            SQL_IS_INTEGER);                  // SQLINTEGER StringLength
   if (ret != SQL_SUCCESS) {
      ivLog->Print ("ERROR: Attempt to set manual-commit mode failed.\n");
      DisplayError (ivLog, ret, SQL_HANDLE_DBC, henv);
      return IVR_FAIL;
   }

   transOpen = true;
   return IVR_SUCCESS;
}
/*SQLList::BeginTrans*/

/* -------------------------------------------------------------------------
   Name:        SQLList::EndTrans

   Purpose:     Ends the current SQL transaction.

   Notes:       - Function should always succeed.  If a transaction has not
                  been opened, it should not be considered a failure when
                  end transaction is called.
                - Now that ODBC is being used, the concept of transactions
                  is different.  Transactions are simply any processing that
                  is done between commits.  By default ODBC is in
                  auto-commit mode.  This function simply sets the commit
                  mode back to auto-commit mode (it may already be there).
   ------------------------------------------------------------------------- */
IV_STATUS SQLList::EndTrans(bool commitData)
{
   SQLRETURN  ret;

   if (!transOpen)
      return IVR_SUCCESS;

   if (commitData)
      ret = SQLEndTran (SQL_HANDLE_ENV, henv, SQL_COMMIT);
   else
      ret = SQLEndTran (SQL_HANDLE_ENV, henv, SQL_ROLLBACK);

   if (ret != SQL_SUCCESS) {
      ivLog->Print ("ERROR: Attempt to %s data failed.\n",
                    (commitData ? "commit" : "rollback"));
      DisplayError (ivLog, ret, SQL_HANDLE_ENV, henv);
   }

   ret = SQLSetConnectAttr (
            hdbc,                             // SQLHDBC    ConnectionHandle
            SQL_ATTR_AUTOCOMMIT,              // SQLINTEGER Attribute
            (SQLPOINTER) SQL_AUTOCOMMIT_ON,   // SQLPOINTER ValuePtr
            SQL_IS_INTEGER);                  // SQLINTEGER StringLength
   if (ret != SQL_SUCCESS) {
      ivLog->Print ("ERROR: Attempt to set auto-commit mode failed.\n");
      DisplayError (ivLog, ret, SQL_HANDLE_DBC, hdbc);
      return IVR_FAIL;
   }

   transOpen = false;
   return IVR_SUCCESS;
}
/*SQLList::EndTrans*/

/* -------------------------------------------------------------------------
   Name:        SQLList::FindSQL
   ------------------------------------------------------------------------- */
SQL *SQLList::FindSQL(const char *requestedName)
{
   string ucName;
   SQL *temp;
   char* value;

   if(requestedName == NULL)
   {
      ivLog->Print("      ERROR: Statement handle parameter empty\n");
      return(NULL);
   }

   // Trim, copy to a string, and make all uppercase characters
   value = TrimStr(requestedName);
   ucName = value;
   delete[] value;
   transform(ucName.begin(), ucName.end(), ucName.begin(), toupper);

   temp = (SQL*)First();
   while(temp)
   {
      // current SQL object's reference name is a match for ucName contents?
      if(ucName.compare(temp->GetReferenceName()) == 0)
      {
         return(temp);
      }
      temp = (SQL*)temp->Suc();
   }
   return(NULL);
}
/*SQList::FindSQL*/

// -------------------------------------------------------------------------
// Name:        SQLList::SaveConnectionString
// -------------------------------------------------------------------------
void SQLList::SaveConnectionString(const char* connectString)
{
   char* value;

   if(connectString)
   {
      value = TrimStr(connectString);
      connectionParms = value;
      delete[] value;
   }
}
//SQLList::SaveConnectionString

// -------------------------------------------------------------------------
// Name:        SQLList::ConnectionStringMatch
// -------------------------------------------------------------------------
bool SQLList::ConnectionStringMatch(const char* compareStr)
{
   if(compareStr)
   {
      return(connectionParms.compare(compareStr) == 0);
   }
   return(false);
}
//SQLList::ConnectionStringMatch

/* -------------------------------------------------------------------------
   Name:        SQLList::SetChildDBHandle
   ------------------------------------------------------------------------- */
void SQLList::SetChildDBHandle(void)
{
   SQL *temp;

   temp = (SQL*)First();
   while (temp) {
      temp->SetDBHandle (hdbc);
      temp = (SQL*)temp->Suc();
   }
}
/*SQList::SetChildDBHandle*/

/* -------------------------------------------------------------------------
   Name:        SQLList::SetChildTransOpen
   ------------------------------------------------------------------------- */
void SQLList::SetChildTransOpen(void)
{
   SQL *temp;

   temp = (SQL*)First();
   while (temp) {
      temp->SetTransOpen (transOpen);
      temp = (SQL*)temp->Suc();
   }
}
/*SQList::SetChildTransOpen*/

/* *************************************************************************
   ************************** SQL Class ************************************
   ************************************************************************* */

/* -------------------------------------------------------------------------
   Name:        SQL::SQL                                       (Constructor)

   Purpose:     Initialize IDAPI.

   Notes:       - May need to look at passing parameters to 'DbiInit'.
   ------------------------------------------------------------------------- */
SQL::SQL(void) : Link()
{
   hdbc      = SQL_NULL_HANDLE;
   hstmt     = SQL_NULL_HANDLE;
   transOpen = false;
   rowCount  = 0;
}
/*SQL::SQL*/

/* -------------------------------------------------------------------------
   Name:        SQL::~SQL                                       (Destructor)
   ------------------------------------------------------------------------- */
SQL::~SQL(void)
{
   EndStmt();
}
/*SQL::~SQL*/

/* -------------------------------------------------------------------------
   Name:        SQL::Lock

   Purpose:     Attempt to lock the current record.

   Notes:       - The 'Fetch' function can lock the record immediately.  May
                  want to support this in the future.

   Questions:   - When does a lock get released?  Does it get released when
                  the cursor moves or do we have to explicitly release the
                  lock?
   ------------------------------------------------------------------------- */
IV_STATUS SQL::Lock()
{
   return IVR_FAIL;
}
/*SQL::Lock*/

/* -------------------------------------------------------------------------
   Name:        SQL::BindParameter                                 (private)
   ------------------------------------------------------------------------- */
IV_STATUS SQL::BindParameter(PIVRINFO pInfo, SQLUSMALLINT pos, SQLParm* sqlParm)
{
   SQLRETURN    ret;

   ret = SQLBindParameter (
         hstmt,                   // Statement handle.
         pos,                     // Parameter number.
         sqlParm->ioType,        // Input output type.
         SQL_C_CHAR,              // Value type.
         sqlParm->parmType,       // Parameter type.
         sqlParm->iColSize,       // Column size.
         sqlParm->iColPrec,       // Decimal digits.
         sqlParm->pszColValue,    // Parameter value pointer.
         sqlParm->iColSize + 1,   // Buffer length.
         &sqlParm->sqlnts);       // StrLen_or_IndPtr.

   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO) {
      DisplayError(pInfo->ivLog, ret, SQL_HANDLE_DBC, hdbc);
      return IVR_FAIL;
   }

   const char* value = pInfo->variables->Get(sqlParm->pszVarName);

   if(value != NULL)
   {
      strcpy (sqlParm->pszColValue, value);
   }
   else
   {
      strcpy (sqlParm->pszColValue, "");
   }
   return IVR_SUCCESS;
}
/*SQL::BindParameter*/

/* -------------------------------------------------------------------------
   Name:        SQL::CallProcedure                                  (public)
   ------------------------------------------------------------------------- */
IV_STATUS SQL::CallProcedure(PIVRINFO pInfo, char* pszStmt, char* pszParams)
{
   SQLParm   *sqlParms = 0;
   int        parmCount;
// SQLCall    sqlCall;
   SQLRETURN  ret;
   IV_STATUS  status;

   if (hdbc == SQL_NULL_HANDLE) {
      pInfo->ivLog->Print("      ERROR: CallProcedure: Not connected to a database\n");
      return IVR_FAIL;
   }

   if (!ParseParamStrings (pszStmt, pszParams, &parmCount, &sqlParms)) {
      delete [] sqlParms;
      return IVR_FAIL;
   }
   // Close any previous statements if still open.
   EndStmt();

   // Allocate a new SQLHDBC. If successful then call SQLConnect using the
   // allocated hdbc and supplied connection information.
   ret = SQLAllocHandle (
            SQL_HANDLE_STMT,            // SQLSMALLINT  HandleType
            hdbc,                       // SQLHANDLE    InputHandle
            &hstmt);                    // SQLHANDLE *  OutputHandle
   if (ret != SQL_SUCCESS) {
      DisplayError(pInfo->ivLog, ret, SQL_HANDLE_STMT, hstmt);
      delete [] sqlParms;
      return IVR_FAIL;
   }

   for (SQLUSMALLINT ii = 0; ii < parmCount; ii++) {
      BindParameter(pInfo, SQLUSMALLINT(ii + 1), &(sqlParms[ii]));
   }

   pInfo->ivLog->Debug ('s', "         <----- BEGIN ----->\n"
                                "%s\n"
                      "         <-----  END  ----->\n",
                 pszStmt);

   ret = SQLExecDirect (
            hstmt,                            // SQLHSTMT    StatementHandle
            (SQLCHAR*)pszStmt,                // SQLCHAR *   StatementText
            SQL_NTS);                         // SQLINTEGER  TextLength

   // If failed to execute statement, free the allocated handle before return.
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO &&
       ret != SQL_NO_DATA)
   {
      DisplayError(pInfo->ivLog, ret, SQL_HANDLE_STMT, hstmt);
      SQLFreeHandle (SQL_HANDLE_STMT, hstmt);
      hstmt = SQL_NULL_HANDLE;
      status = IVR_FAIL;
   } else {
      for (int ii = 0; ii < parmCount; ii++) {
         if (sqlParms[ii].sqlnts == SQL_NULL_DATA)
            *(sqlParms[ii].pszColValue) = 0;
         else
            sqlParms[ii].pszColValue[sqlParms[ii].sqlnts] = 0;
         pInfo->ivLog->Print ("%d: '%s' = '%s'\n", ii,
                       sqlParms[ii].pszVarName,
                       sqlParms[ii].pszColValue);
         if (sqlParms[ii].ioType != SQL_PARAM_INPUT) {
            if(!pInfo->variables->Set(sqlParms[ii].pszVarName, sqlParms[ii].pszColValue))
            {
               pInfo->ivLog->Debug ('D', "Failed to set variable '%s'.\n",
                             sqlParms[ii].pszVarName);
               delete [] sqlParms;
               return IVR_FAIL;
            }
         }
      }
      status = IVR_SUCCESS;
   }

   if (ret == SQL_SUCCESS_WITH_INFO)
      EATERRORS(pInfo->ivLog, ret, SQL_HANDLE_STMT, hdbc);

   delete [] sqlParms;
   return status;
}
/*SQL::CallProcedure*/

/* -------------------------------------------------------------------------
   Name:        SQL::ParseParamStrings                             (private)
   ------------------------------------------------------------------------- */
bool SQL::ParseParamStrings (
      char     *pszStmt,
      char     *pszParams,
      int      *parmCount,
      SQLParm **sqlParms)
{
   char *pszMark;

   *parmCount = 0;
   pszMark = strchr (pszStmt, '?');
   while (pszMark && *pszMark)
   {
      (*parmCount)++;
      pszMark = strchr (pszMark + 1, '?');
   }

   if (*parmCount > 0)
      *sqlParms = new SQLParm [*parmCount];

   char *pszSOL;
   char *pszEOL;
   int   ii;

   ii = 0;
   pszSOL = pszParams;
   pszEOL = strchr (pszSOL, '\n');
   while (pszEOL && ii < *parmCount) {
      // Replace the NewLine character with a NUL (to NUL-terminate the string)
      // before calling the ParseParamString() function.
      *pszEOL = 0;
      if (!ParseParamString (pszSOL, &((*sqlParms)[ii])))
         return false;
      ii++;
      pszSOL = pszEOL + 1;
      // Restore the NewLine character (replaced with NUL above).
      *pszEOL = '\n';
      pszEOL = strchr (pszSOL, '\n');
   }
   if (ii < *parmCount)
      return false;
   return true;
}
/*SQL::ParseParamStrings*/

/* -------------------------------------------------------------------------
   Name:        SQL::ParseParamString                              (private)
   ------------------------------------------------------------------------- */
bool SQL::ParseParamString (
      char    *pszValue,
      SQLParm *sqlParm)
{
   char    *pszStart = pszValue;
   char    *pszEnd;
   char     c;
   int      iFieldNo = 1;
   int      iLen;

   while (iFieldNo < 6 && pszStart)
   {
      pszEnd = strchr (pszStart, ',');
      if (iFieldNo < 5 && (!pszEnd || !*pszEnd)) {
//         ivLog->Debug ('S', "Missing fields in parameter string.\n");
         return false;
      }
      if (!pszEnd)
         pszEnd = strchr (pszStart, '\0');

      iLen = pszEnd - pszStart;

      switch (iFieldNo)
      {
      case 1:
         sqlParm->pszVarName = new char [iLen + 1];
         // *** HACK *** This function used to be in the BRD_XXX_CODE.DLL
         // (where it didn't belong) and I needed to have two @ characters
         // to prevent the variables from being parsed before they arrived
         // in this function.  Now that I've moved the code to it's right
         // location, the following 4 lines of code are not needed except
         // to support systems built before I moved the code here.
         if (*pszStart == '@' && *(pszStart + 1) == '@') {
            pszStart++;
            iLen--;
         }
         strncpy (sqlParm->pszVarName, pszStart, iLen);
         sqlParm->pszVarName[iLen] = 0;
         break;

      case 2:
         if (iLen == 5 && !strncmp (pszStart, "INOUT", 5))
            sqlParm->ioType = SQL_PARAM_INPUT_OUTPUT;
         else if (iLen == 2 && !strncmp (pszStart, "IN", 2))
            sqlParm->ioType = SQL_PARAM_INPUT;
         else if (iLen == 3 && !strncmp (pszStart, "OUT", 3))
            sqlParm->ioType = SQL_PARAM_OUTPUT;
         break;

      case 3:
         if (iLen == 6 && !strncmp (pszStart, "NUMBER", 6))
            sqlParm->parmType = SQL_NUMERIC;
         else if (iLen == 7 && !strncmp (pszStart, "VARCHAR", 7))
            sqlParm->parmType = SQL_VARCHAR;
         break;

      case 4:
         c = *pszEnd;
         *pszEnd = 0;
         sqlParm->iColSize = atoi (pszStart);
         if (sqlParm->iColSize > 0)
            sqlParm->pszColValue = new char [sqlParm->iColSize + 1];
         *pszEnd = c;
         break;

      case 5:
         c = *pszEnd;
         *pszEnd = 0;
         sqlParm->iColPrec = (SQLSMALLINT) atoi (pszStart);
         *pszEnd = c;
      }

      iFieldNo++;
      if (pszEnd && *pszEnd)
         pszStart = pszEnd + 1;
      else
         pszStart = 0;
   }
   return true;
}
/*StepSQLCall::ParseParamString*/

/* -------------------------------------------------------------------------
   Name:        SQL::BeginStmt

   Purpose:     Opens and executes an SQL statement.

   Notes:       - Need to add check to ensure a cursor is not already
                  opened.
   ------------------------------------------------------------------------- */
IV_STATUS SQL::BeginStmt(PIVRINFO pInfo, const char* szQry)
{
   SQLRETURN   ret;
   IV_STATUS   status;
   char* szFullQry;

   if(hdbc == SQL_NULL_HANDLE)
   {
      pInfo->ivLog->Print ("ERROR: BeginStmt: Not connected to a database.\n");
      return(IVR_FAIL);
   }

   rowCount = 0;

   // Close any previous statements if still open.
   EndStmt();

   szFullQry = pInfo->variables->Parse(szQry);
   if(szFullQry == NULL)
   {
      pInfo->ivLog->Print ("ERROR: Parsing SQL statement failed.\n");
      pInfo->ivLog->Debug ('s', "         <----- BEGIN ----->\n"
                                   "%s\n"
                         "         <-----  END  ----->\n",
                    szQry ? szQry : "");
      return(IVR_FATAL);
   }
   pInfo->ivLog->Debug ('s', "         <----- BEGIN ----->\n"
                                "%s\n"
                      "         <-----  END  ----->\n",
                 szFullQry);

   // Allocate a new SQLHDBC. If successful then call SQLConnect using the
   // allocated hdbc and supplied connection information.
   ret = SQLAllocHandle (
            SQL_HANDLE_STMT,            // SQLSMALLINT  HandleType
            hdbc,                       // SQLHANDLE    InputHandle
            &hstmt);                    // SQLHANDLE *  OutputHandle
   if (ret != SQL_SUCCESS) {
      DisplayError(pInfo->ivLog, ret, SQL_HANDLE_STMT, hstmt);
      delete[] szFullQry;
      return(IVR_FAIL);
   }

   ret = SQLExecDirect (
            hstmt,                      // SQLHSTMT    StatementHandle
            (SQLCHAR*)szFullQry,        // SQLCHAR *   StatementText
            SQL_NTS);                   // SQLINTEGER  TextLength
   delete[] szFullQry;

   // If failed to execute statement, free the allocated handle before return.
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO &&
       ret != SQL_NO_DATA)
   {
      pInfo->ivLog->Print ("Error code %hd returned from ODBC driver\n", ret);
      DisplayError(pInfo->ivLog, ret, SQL_HANDLE_STMT, hstmt);
      SQLFreeHandle (SQL_HANDLE_STMT, hstmt);
      hstmt = SQL_NULL_HANDLE;
      status = IVR_FAIL;
   } else
      status = IVR_SUCCESS;

   if (ret == SQL_SUCCESS_WITH_INFO)
   {
      pInfo->ivLog->Print("         Success with info\n");
      EATERRORS(pInfo->ivLog, ret, SQL_HANDLE_STMT, hdbc);
   }

   return status;
}
/*SQL::BeginStmt*/

/* -------------------------------------------------------------------------
   Name:        SQL::EndStmt

   Purpose:     Closes a previously opened SQL statement.  Free memory that
                was used to store the record.


   Questions:   - Should this function fail if a statement hasn't been
                  opened?
   ------------------------------------------------------------------------- */
IV_STATUS SQL::EndStmt(void)
{
   if (hdbc == SQL_NULL_HANDLE) {
      return IVR_FAIL;
   }

   if (hstmt != SQL_NULL_HANDLE) {
      SQLFreeHandle (SQL_HANDLE_STMT, hstmt);
      hstmt = SQL_NULL_HANDLE;
   }

   return IVR_SUCCESS;
}
/*SQL::EndStmt*/

/* -------------------------------------------------------------------------
   Name:        SQL::HaveResultSet
   ------------------------------------------------------------------------- */
bool SQL::HaveResultSet(void)
{
   SQLRETURN     ret;
   SQLSMALLINT   numCols;

   ret = SQLNumResultCols (hstmt, &numCols);
   if (ret != SQL_SUCCESS) {
//      DisplayError(ivLog, ret, SQL_HANDLE_STMT, hstmt);
      return IVR_FAIL;
   }

   if (!numCols)
      return false;

   return true;
}
/*SQL::HaveResultSet*/

/* -------------------------------------------------------------------------
   Name:        SQL::Fetch

   Purpose:     Retrieves one record from the opened SQL statement.

   Notes:       - May want to set the translation mode prior to retrieving
                  field descriptions.
                - Uses 'VarSet' to assign fields fetched to variables->
                  Variables are created using the default method.
   ------------------------------------------------------------------------- */
IV_STATUS SQL::Fetch(PIVRINFO pInfo)
{
   SQLRETURN     ret;
   SQLSMALLINT   numCols;
   SQLUSMALLINT  col;
   SQLCHAR       value[1024];
   string        entireValue;   // used to concatenate blocks of text larger than 1024
   SQLCHAR       colName[81];
   SQLSMALLINT   colNameLen;
   SQLSMALLINT   colDataType;
   SQLULEN       colSize;
   SQLSMALLINT   colDecimal;
   SQLSMALLINT   colNullable;
   SQLLEN        iStrLenOrIndPtr;

   if (hdbc == SQL_NULL_HANDLE) {
      pInfo->ivLog->Print ("ERROR: No database connect available.\n");
      return IVR_FAIL;
   }

   if (hstmt == SQL_NULL_HANDLE) {
      pInfo->ivLog->Print ("ERROR: No SQL statement defined.\n");
      return IVR_FAIL;
   }

   ret = SQLNumResultCols (hstmt, &numCols);
   if (ret != SQL_SUCCESS) {
      DisplayError(pInfo->ivLog, ret, SQL_HANDLE_STMT, hstmt);
      return IVR_FAIL;
   }
   if (numCols == 0) {
      pInfo->ivLog->Print("ERROR: Can only fetch records after select statements.\n");
      return IVR_FAIL;
   }

   ret = SQLFetch (hstmt);
   if (ret == SQL_SUCCESS)
   {
      rowCount++;

      for (col = 1; col <= numCols; ++col)
      {
         ret = SQLDescribeCol (
                 hstmt,                 // SQLHSTMT     StatementHandle
                 col,                   // SQLSMALLINT  ColumnNumber
                 colName,               // SQLCHAR     *ColumnName
                 80,                    // SQLSMALLINT  BufferLength
                 &colNameLen,           // SQLSMALLINT *NameLengthPtr
                 &colDataType,          // SQLSMALLINT *DataTypePtr
                 &colSize,              // SQLUINTEGER *ColumnSizePtr
                 &colDecimal,           // SQLSMALLINT *DecimalDigitsPtr
                 &colNullable);         // SQLSMALLINT *NullablePtr

         if (ret != SQL_SUCCESS)
            DisplayError(pInfo->ivLog, ret, SQL_HANDLE_STMT, hstmt);

         strcpy ((char*)value, (char*)"@");
         strcat ((char*)value, (char*)colName);
         strcpy ((char*)colName, (char*)value);
         *value = 0;

         ret = SQLGetData (
                  hstmt,                // SQLHSTMT     StatementHandle
                  col,                  // SQLUSMALLINT ColumnNumber
                  SQL_C_CHAR,           // SQLSMALLINT  TargetType
                  value,                // SQLPOINTER   TargetValuePtr
                  1024,                 // SQLINTEGER   BufferLength,
                  &iStrLenOrIndPtr);    // SQLINTEGER * StrLen_or_IndPtr

		 if((ret != SQL_SUCCESS) && (ret != SQL_SUCCESS_WITH_INFO))
            DisplayError(pInfo->ivLog, ret, SQL_HANDLE_STMT, hstmt);

         // Copy the data returned so far to entireValue string
         entireValue = (char*) value;

         switch (iStrLenOrIndPtr)
         {
            case SQL_NO_TOTAL:
            case SQL_NULL_DATA:
               *value = 0;
               break;
            default:
               if(iStrLenOrIndPtr > 1023)
               {
                  int remainingCount = (iStrLenOrIndPtr - 1023) + 1;   // account for null terminator in previous call
				                                                       // and the one we'll receive in the next call
                  SQLCHAR* value2 = new SQLCHAR[remainingCount];

                  ret = SQLGetData (
                        hstmt,                // SQLHSTMT     StatementHandle
                        col,                  // SQLUSMALLINT ColumnNumber
                        SQL_C_CHAR,           // SQLSMALLINT  TargetType
                        value2,               // SQLPOINTER   TargetValuePtr
                        remainingCount,       // SQLINTEGER   BufferLength,
                        &iStrLenOrIndPtr);    // SQLINTEGER * StrLen_or_IndPtr

                  if((ret != SQL_SUCCESS) && (ret != SQL_SUCCESS_WITH_INFO))
                     DisplayError(pInfo->ivLog, ret, SQL_HANDLE_STMT, hstmt);

				  // Copy the rest of the data to entireValue string
                  entireValue += (char*) value2;
               }
               break;
         }

         if(!pInfo->variables->Set((char*)colName, entireValue.c_str()))
         {
            pInfo->ivLog->Print("Failed to set variable '%s'.\n", colName);
            return IVR_FAIL;
         }
      }

   }
	else if (ret == SQL_NO_DATA)
	{
      return IVR_FAIL;
   }
	else
	{
      DisplayError(pInfo->ivLog, ret, SQL_HANDLE_STMT, hstmt);
      return IVR_FAIL;
   }

   return IVR_SUCCESS;
}
/*SQL::Fetch*/

IV_STATUS SQL::SetReferenceName(const char* value)
{
   string ucName;
   if(value == NULL)
   {
      return(IVR_FATAL);
   }

   // Make a string of uppercase characters
   ucName = value;
   transform(ucName.begin(), ucName.end(), ucName.begin(), toupper);

   refName = ucName;
   return(IVR_SUCCESS);
}
//SQL::SetReferenceName

/* *************************************************************************
   *********************** Support Functions *******************************
   ************************************************************************* */

/* -------------------------------------------------------------------------
   Name:        DisplayError
   ------------------------------------------------------------------------- */
void DisplayError (
   Log        *ivLog,
   SQLRETURN   /*nResult*/,
   SWORD       fHandleType,
   SQLHANDLE   handle)
{
   SQLRETURN   ret;                                   //
   SQLINTEGER  dwErrCode;                             // Native Error code
   SQLSMALLINT sMsgNum;                               //
   SQLSMALLINT wErrMsgLen;                            //
   UCHAR       szErrState[SQL_SQLSTATE_SIZE + 1];     // SQL error state string
   UCHAR       szErrText[SQL_MAX_MESSAGE_LENGTH + 1]; // SQL error text string

   sMsgNum = 1;

   ret = SQLGetDiagRec (
            fHandleType,             // SQLSMALLINT   HandleType
            handle,                  // SQLHANDLE     Handle
            sMsgNum,                 // SQLSMALLINT   RecNumber
            szErrState,              // SQLCHAR *     Sqlstate
            &dwErrCode,              // SQLINTEGER *  NativeError
            szErrText,               // SQLCHAR *     MessageText
            SQL_MAX_MESSAGE_LENGTH,  // SQLSMALLINT   BufferLength
            &wErrMsgLen);            // SQLSMALLINT * TextLength

   while (ret != SQL_NO_DATA && ret != SQL_ERROR && ret != SQL_INVALID_HANDLE)
   {
      ++sMsgNum;

      ivLog->Print ("\tSQL Error State:   %s\n"
                   "\tNative Error Code: %lX\n"
                   "\tODBC Error:        %s\n",
                   szErrState, dwErrCode, szErrText);

      ret = SQLGetDiagRec (
               fHandleType,             // SQLSMALLINT   HandleType
               handle,                  // SQLHANDLE     Handle
               sMsgNum,                 // SQLSMALLINT   RecNumber
               szErrState,              // SQLCHAR *     Sqlstate
               &dwErrCode,              // SQLINTEGER *  NativeError
               szErrText,               // SQLCHAR *     MessageText
               SQL_MAX_MESSAGE_LENGTH,  // SQLSMALLINT   BufferLength
               &wErrMsgLen);            // SQLSMALLINT * TextLength
   }
}
/*DisplayError*/


/* -------------------------------------------------------------------------
   Name:        EatErrors
   ------------------------------------------------------------------------- */
void EatErrors (
   Log        *, // ivLog,
   SQLRETURN   /*nResult*/,
   SWORD       fHandleType,
   SQLHANDLE   handle)
{
   SQLRETURN   ret;                                   //
   SQLINTEGER  dwErrCode;                             // Native Error code
// SQLINTEGER  sMsgNum;                               //
   SQLSMALLINT sMsgNum;                               //
   SQLSMALLINT wErrMsgLen;                            //
   UCHAR       szErrState[SQL_SQLSTATE_SIZE + 1];     // SQL error state string
   UCHAR       szErrText[SQL_MAX_MESSAGE_LENGTH + 1]; // SQL error text string

   ret = SQL_SUCCESS;
   sMsgNum = 1;
   while (ret != SQL_NO_DATA && ret != SQL_ERROR && ret != SQL_INVALID_HANDLE) {
      ret = SQLGetDiagRec (
               fHandleType,             // SQLSMALLINT   HandleType
               handle,                  // SQLHANDLE     Handle
               sMsgNum,                 // SQLSMALLINT   RecNumber
               szErrState,              // SQLCHAR *     Sqlstate
               &dwErrCode,              // SQLINTEGER *  NativeError
               szErrText,               // SQLCHAR *     MessageText
               SQL_MAX_MESSAGE_LENGTH,  // SQLSMALLINT   BufferLength
               &wErrMsgLen);            // SQLSMALLINT * TextLength
      ++sMsgNum;
   }
}

/* -------------------------------------------------------------------------
   Name:        SQLParm::SQLParm
   ------------------------------------------------------------------------- */
SQLParm::SQLParm()
{
   pszVarName  = 0;
   pszColValue = 0;
   iColSize    = 0;
   iColPrec    = 0;
   ioType      = 0;
   parmType    = 0;
   sqlnts      = SQL_NTS;
}
/*SQLParm::SQLParm*/

/* -------------------------------------------------------------------------
   Name:        SQLParm::~SQLParm
   ------------------------------------------------------------------------- */
SQLParm::~SQLParm()
{
   delete [] pszVarName;
   delete [] pszColValue;
}
/*SQLParm::~SQLParm*/
#if 0
/* -------------------------------------------------------------------------
   Name:        SQLCall::SQLCall
   ------------------------------------------------------------------------- */
SQLCall::SQLCall()
{
   pszStatement = 0;
   iParmCount = 0;
   sqlParms = 0;
}
/*SQLCall::SQLCall*/

/* -------------------------------------------------------------------------
   Name:        SQLCall::~SQLCall
   ------------------------------------------------------------------------- */
SQLCall::~SQLCall()
{
   delete [] pszStatement;
   delete [] sqlParms;
}
/*SQLCall::~SQLCall*/
#endif
#if 0
/* -------------------------------------------------------------------------
   Name:        main
   ------------------------------------------------------------------------- */
void main (int argc, char *argv[])
{
   IV_STATUS   status;
   char        s[1024];
   SQLList    *sqlList;
   SQL        *stmt;
   ConsoleLog  ivLog;
   VarList     globList;
   int         iRefNo;

   if (argc == 1) {
      ivLog.Print ("SYNTAX: jr1 <dsn>\n");
      return;
   }

   sqlList = new SQLList (&ivLog);
   if (!sqlList || !sqlList->Okay()) {
      ivLog.Print ("ERROR: Failed to attach to database.  Check DSN.\n");
      delete sqlList;
      return;
   }

   if (sqlList->Connect (argv[1]) != IVR_SUCCESS) {
      delete sqlList;
      return;
   }

   sqlList->DetermineTransCapability();

   ivLog.Print ("Type 'exit' to quit.\n");
   ivLog.Print ("SQL> ");
   gets (s);
   for (;;)
   {
      if (!stricmp (s, "exit") || !stricmp (s, "quit"))
         break;

      if (!stricmp (s, "begin")) {
         ivLog.Print ("Begin transaction.\n");
         sqlList->BeginTrans();
      }

      else if (!stricmp (s, "commit")) {
         if (sqlList->EndTrans (true) == IVR_SUCCESS)
            ivLog.Print ("Commit complete.\n");
         else
            ivLog.Print ("Commit failed.\n");
      }

      else if (!stricmp (s, "rollback")) {
         if (sqlList->EndTrans (false) == IVR_SUCCESS)
            ivLog.Print ("Rollback complete.\n");
         else
            ivLog.Print ("Rollback failed.\n");
      }

      else if (!strnicmp (s, "fetch", 5)) {
         iRefNo = atoi (s+5);
         stmt = sqlList->FindSQL (iRefNo);
         if (stmt) {
            status = stmt->Fetch (0, &globList);
            if (status != IVR_SUCCESS)
               ivLog.Print ("Nothing more to fetch.\n");
         }
      }

      else if (!strnicmp (s, "close", 5)) {
         iRefNo = atoi (s+5);
         stmt = sqlList->FindSQL (iRefNo);
         if (stmt)
            delete stmt;
         ivLog.Print ("Closed.\n");
      }

      else {
         iRefNo = atoi (s);
         ivLog.Print ("iRefNo = %d\n", iRefNo);
         stmt = sqlList->FindSQL (iRefNo);
         if (stmt)
            delete stmt;
         stmt = new SQL (&ivLog);
         stmt->SetReferenceNo (iRefNo);
         stmt->SetDBHandle (sqlList->GetDBHandle());
         sqlList->AddLast (stmt);
         if ((status = stmt->BeginStmt (0, &globList, s + 2)) == IVR_SUCCESS) {
            if (stmt->HaveResultSet())
               ivLog.Print ("Data ready to be fetched.\n");
            else
               ivLog.Print ("Command complete.  Don't forget to 'close'.\n");
         }
      }
      ivLog.Print ("SQL> ");
      gets (s);
   }

   sqlList->Disconnect();
   delete sqlList;
}
/*main*/
#endif

